# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Supplement for chicago taxi pipeline example with specifics schema modification.
This module file will be used in the custom schema curation component.
"""

import tensorflow_data_validation as tfdv

# TFX schema curation component will call this function.


def schema_fn(schema):
  """modifies the infered schema.
  Args:
    schema:schema generated by SchemaGen component of tfx
  """
  #changing "tips" into optional feature
  feature = tfdv.get_feature(schema, 'tips')
  feature.presence.min_fraction = 0.9

  return schema
